﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015.Environment
{
    public class WorldGenerator
    {

        int recursionCount = 0;
        int recursionMax = 100;

        int width, height;
        Random rand;

        int[,] tiles;

        /// <summary>
        /// This is the constructor for a world generator. This is used
        /// to generate a new world of a given width and height. The height
        /// of 300 is strongly recommended.
        /// </summary>
        /// <param name="w">The width of the map</param>
        /// <param name="h">The height of the map</param>
        public WorldGenerator(int w, int h = 300)
        {
            this.width = w;
            this.height = h;
        }

        /// <summary>
        /// This metod sets the PRNG seed of the random object
        /// </summary>
        /// <param name="seed">The seed to set</param>
        private void SetSeed(int seed)
        {
            rand = new Random(seed);
        }

        /// <summary>
        /// This method is used to actually generate and return the new world.
        /// </summary>
        /// <param name="seed">The seed of the world to generate. -1 means to use the current time as a seed</param>
        /// <returns>The array of block id's representing the generated world</returns>
        public int[,] Generate(int seed = -1)
        {
            //Set the PRNG seed
            if (seed == -1)
            {
                SetSeed(DateTime.Now.Ticks.GetHashCode());
            }

            //Create the tile array
            tiles = new int[width, height];

            //Run Joe's first pass algorithm (slightly tweaked)
            JoeFirstPass();

            JoeSecondPass();

            //Run Joe's Ore Placement algorithm
            JoeOrePlacement();

            //Run Cellular Automata on the ores to make veins look more like clumps
            OreCellularAutomata();
            OreCellularAutomata();

            return tiles;
        }

        /// <summary>
        /// This method will generate the initial block types of the map, ignoring
        /// ore placement. This comes later.
        /// </summary>
        public void JoeFirstPass()
        {
            int heightCounter = 1;

            //B Height 1: grass top layer
            for (int i = 0; i < width; i++)
            {
                tiles[i, 0] = 1;
            }


            //B Height 1 - 10%: dirt 100%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter + (height / 10); j++)
                {
                    tiles[i, j] = Block.DIRT;
                }
            }

            heightCounter = (height / 10);

            //B Height 10%-15% : dirt 50%, stone 50%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter + (height / 20); j++)
                {
                    int a = rand.Next(0, 10);
                    tiles[i, j] = a < 5 ? Block.DIRT : Block.STONE;
                }
            }

            heightCounter += (height / 20);

            //B Height 15% - 30%: dirt 25%, stone 75%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter * 2; j++)
                {
                    int a = rand.Next(0, 100);
                    if (a < 25)
                    {
                        tiles[i, j] = Block.DIRT;
                    }
                    else
                    {
                        tiles[i, j] = Block.STONE;
                    }
                }
            }

            heightCounter *= 2;

            //B Height 30 - 60:  dirt 15%, stone 85%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter * 2; j++)
                {
                    int a = rand.Next(0, 100);

                    if (a < 15)
                    {
                        tiles[i, j] = Block.DIRT;
                    }
                    else
                    {
                        tiles[i, j] = Block.STONE;
                    }
                }
            }

            heightCounter *= 2;

            //B Height 60% - 70%: granite 20%, stone 80%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter + (height / 10); j++)
                {
                    int a = rand.Next(0, 10);
                    if (a < 1)
                    {
                        tiles[i, j] = Block.GRANITE;
                    }
                    else
                    {
                        tiles[i, j] = Block.STONE;
                    }

                }
            }

            heightCounter += (height / 10);

            //B Height 70% - 80%: granite 20%, stone 80%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter + (height / 10); j++)
                {
                    int a = rand.Next(0, 10);
                    if (a < 2)
                    {
                        tiles[i, j] = Block.GRANITE;
                    }
                    else
                    {
                        tiles[i, j] = Block.STONE;
                    }
                }
            }

            heightCounter += (height / 10);

            //B Height 80% - 90%: stone 70%, Granite 30%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < heightCounter + (height / 10); j++)
                {
                    int a = rand.Next(0, 10);
                    if (a < 7)
                    {
                        tiles[i, j] = Block.STONE;
                    }
                    else
                    {
                        tiles[i, j] = Block.GRANITE;
                    }
                }
            }

            heightCounter += (height / 10);

            //B Height 90% - 100%: stone 75%, granite 25%
            for (int i = 0; i < width; i++)
            {
                for (int j = heightCounter; j < height; j++)
                {
                    int a = rand.Next(0, 100);
                    if (a < 75)
                    {
                        tiles[i, j] = Block.STONE;
                    }
                    else
                    {
                        tiles[i, j] = Block.GRANITE;
                    }

                }
            }



            //B Height 293 - 300: bedrock 100%
            for (int i = 0; i < width; i++)
            {

                tiles[i, height - 1] = Block.BEDROCK;

            }
        }


        /// <summary>
        /// Adds veins of dirt, stone and bedrock
        /// </summary>
        public void JoeSecondPass()
        {
            GenerateResourceVein(Block.STONE_SPAWN, true);
            GenerateResourceVein(Block.DIRT_SPAWN, true);
            GenerateResourceVein(Block.GRANITE_SPAWN, true);
            GenerateResourceVein(Block.AIR_SPAWN, true);
        }

        /// <summary>
        /// This method is used to place ores in the generated map.
        /// </summary>
        public void JoeOrePlacement()
        {
            //Each one of these calls will spawn in a new resource, based on the spawn data passed
            GenerateResourceVein(Block.TIN_SPAWN);
            GenerateResourceVein(Block.COPPER_SPAWN);
            GenerateResourceVein(Block.BAUXITE_SPAWN);
            GenerateResourceVein(Block.GOLD_SPAWN, true);
            GenerateResourceVein(Block.MITHRIL_SPAWN);
            GenerateResourceVein(Block.ADAMANTIUM_SPAWN);
            GenerateResourceVein(Block.COBALT_SPAWN, true);
            GenerateResourceVein(Block.URANIUM_SPAWN);
            GenerateResourceVein(Block.PLUTONIUM_SPAWN);
            GenerateResourceVein(Block.UNOBTANIUM_SPAWN);

            //Ores are done, now for Gems
            GenerateResourceVein(Block.SAPPHIRE_SPAWN);
            GenerateResourceVein(Block.EMERALD_SPAWN);
            GenerateResourceVein(Block.RUBY_SPAWN);

            GenerateResourceVein(Block.COPPER_SPAWN_DIRT, true);
            GenerateResourceVein(Block.TIN_SPAWN_DIRT, true);


        }

        /// <summary>
        /// This method will generate a resource vein based on the spawn data provided.
        /// </summary>
        /// <param name="data">The spawn data to generate from</param>
        private void GenerateResourceVein(SpawnData data, bool directionWeighting = false)
        {
            recursionCount = 0;
            for (int i = 0; i < data.depositCount; i++)
            {
                int x = rand.Next(width);
                int y = rand.Next(data.minDepth, data.maxDepth != -1 ? data.maxDepth : height);

                if (tiles[x, y] == data.hostType)
                {
                    SetTile(x, y, data.blockID);
                    if (directionWeighting)
                    {
                        GenerateVeinRandomWalk(x, y, data.blockID, data.depositJoinRateW, data.depositJoinRateE, data.depositJoinRateN, data.depositJoinRateS, data.hostType);
                    }
                    else
                    {
                        GenerateVeinRandomWalk(x, y, data.blockID, data.depositJoinRate, data.hostType);
                    }
                }
            }
        }


        /// <summary>
        /// This method is used to actually place the vein of resources
        /// </summary>
        /// <param name="x">The x position to start the vein</param>
        /// <param name="y">The y position to start the vein</param>
        /// <param name="blockType">The block type of the vein</param>
        /// <param name="expansionChance">The chance of expanding the vein </param>
        /// <param name="hostType">The host block type of the resource</param>
        private void GenerateVeinRandomWalk(int x, int y, int blockType, int expansionChance, int hostType)
        {
            if (recursionCount > recursionMax) return;
            recursionCount++;
            /**
             * This method performs a basic recursive random walk to place a vein of resources.
             * Each of these if statements checks in the 4 cardinal directions of the block, so 
             * check suitablity to continue the walk before rolling to see if the PRNG permits 
             * the walk.
             */
            if (GetTile(x - 1, y) == hostType)
            {
                if (rand.Next(0, 100) < expansionChance)
                {
                    SetTile(x - 1, y, blockType);
                    GenerateVeinRandomWalk(x - 1, y, blockType, expansionChance, hostType);
                }

            }

            if (GetTile(x, y - 1) == hostType)
            {
                if (rand.Next(0, 100) < expansionChance)
                {
                    SetTile(x, y - 1, blockType);
                    GenerateVeinRandomWalk(x, y - 1, blockType, expansionChance, hostType);
                }

            }

            if (GetTile(x + 1, y) == hostType)
            {
                if (rand.Next(0, 100) < expansionChance)
                {
                    SetTile(x + 1, y, blockType);
                    GenerateVeinRandomWalk(x + 1, y, blockType, expansionChance, hostType);
                }

            }

            if (GetTile(x, y + 1) == hostType)
            {
                if (rand.Next(0, 100) < expansionChance)
                {
                    SetTile(x, y + 1, blockType);
                    GenerateVeinRandomWalk(x, y + 1, blockType, expansionChance, hostType);
                }
            }
        }

        /// <summary>
        /// Specifies the spawn rates for each direction
        /// </summary>
        /// <param name="x"> original block to walk from</param>
        /// <param name="y"> ^^ </param>
        /// <param name="blockType"></param>
        /// <param name="expansionChanceW"> west expansion</param>
        /// <param name="expansionChanceE"> east expansion</param>
        /// <param name="expansionChanceN"> north expansion</param>
        /// <param name="expansionChanceS"> south expansion</param>
        /// <param name="hostType"> </param>
        private void GenerateVeinRandomWalk(int x, int y, int blockType, int expansionChanceW, int expansionChanceE, int expansionChanceN, int expansionChanceS, int hostType)
        {
            if (recursionCount > recursionMax) return;
            recursionCount++;
            if (GetTile(x - 1, y) == hostType)
            {
                if (rand.Next(0, 100) < expansionChanceW)
                {
                    SetTile(x - 1, y, blockType);
                    GenerateVeinRandomWalk(x - 1, y, blockType, expansionChanceW, expansionChanceE, expansionChanceN, expansionChanceS, hostType);
                }

            }

            if (GetTile(x, y - 1) == hostType)
            {
                if (rand.Next(0, 100) < expansionChanceN)
                {
                    SetTile(x, y - 1, blockType);
                    GenerateVeinRandomWalk(x, y - 1, blockType, expansionChanceW, expansionChanceE, expansionChanceN, expansionChanceS, hostType);
                }

            }

            if (GetTile(x + 1, y) == hostType)
            {
                if (rand.Next(0, 100) < expansionChanceE)
                {
                    SetTile(x + 1, y, blockType);
                    GenerateVeinRandomWalk(x + 1, y, blockType, expansionChanceW, expansionChanceE, expansionChanceN, expansionChanceS, hostType);
                }

            }

            if (GetTile(x, y + 1) == hostType)
            {
                if (rand.Next(0, 100) < expansionChanceS)
                {
                    SetTile(x, y + 1, blockType);
                    GenerateVeinRandomWalk(x, y + 1, blockType, expansionChanceW, expansionChanceE, expansionChanceN, expansionChanceS, hostType);
                }
            }
        }

        /// <summary>
        /// This method is a basic implementation of Cellular Automata to fill in gaps in ore veins.
        /// This method will check to see if the 4 direct neighbours of a block are the same, and if so,
        /// will set the block to the same type.
        /// </summary>
        private void OreCellularAutomata()
        {
            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {

                    if (GetTile(x - 1, y) == GetTile(x + 1, y) && GetTile(x - 1, y) == GetTile(x, y - 1) && GetTile(x, y - 1) == GetTile(x, y + 1))
                    {
                        int tile = GetTile(x - 1, y);
                        if (tile == Block.DIRT || tile == Block.BEDROCK || tile == Block.STONE) continue;

                        SetTile(x, y, tile);
                    }
                }
            }
        }

        /// <summary>
        /// This method is used to safely set a tile, as it will normalise the x coord and perform range
        /// checks on the y coordinate
        /// </summary>
        /// <param name="x">The x coordinate of the tile to set</param>
        /// <param name="y">The y coordinate of the tile to set</param>
        /// <param name="blockType">The block type to set</param>
        private void SetTile(int x, int y, int blockType)
        {
            //Normalise the x to be within the bounds [0, width]
            x = Normalise(x, 0, width);

            //Range check the y coordinate
            if (y < 0 || y >= height) return;

            //Checks passed, set the block
            tiles[x, y] = blockType;
        }

        /// <summary>
        /// This method is used to safely get a tile from the tiles array. It is safe as it will validate
        /// the coordinate passed to it by normalising the x value and range checking the y value.
        /// </summary>
        /// <param name="x">The x coordinate to retrieve</param>
        /// <param name="y">The y coordinate to retrieve</param>
        /// <returns>The retrieved block</returns>
        private int GetTile(int x, int y)
        {
            //If the y range check fails, return Air.
            if (y < 0 || y >= height) return Block.AIR;

            //Normalise the x value
            x = Normalise(x, 0, width);

            //Return the tile
            return tiles[x, y];
        }

        /// <summary>
        /// This method will Normalise a value (input) between min and max
        /// </summary>
        /// <param name="input">The value to normalise</param>
        /// <param name="min">The minimum value</param>
        /// <param name="max">The maximum value</param>
        /// <returns>The normalized value</returns>
        public int Normalise(int input, int min, int max)
        {
            //while (input >= max) input -= (max - min);
            //while (input < min) input += (max - min);
            input %= max;
            if (input < min) input += max;
            return input;
        }
    }
}
