﻿using GGJ_2015.Environment;
using GGJ_2015.Scenes;
using GGJ_2015.Util;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015
{
    public class Block : GameObject
    {
        /// <summary>
        /// This is used to toggle the drawing of the gradient. For debug purposes.
        /// </summary>
        private const bool DRAW_GRADIENT = true;

        /// <summary>
        /// These are the block IDs. The are unique and each one is used to denot a different resource/block type
        /// </summary>
        public const int AIR = -1, GRASS = 1, DIRT = 2, STONE = 3, GRANITE = 4, BEDROCK = 5, TIN = 6, COPPER = 7, BAUXITE = 8, GOLD = 9, MITHRIL = 10;
        public const int ADAMANTIUM = 11, COBALT = 12, URANIUM = 13, PLUTONIUM = 14, UNOBTANIUM = 15, SAPPHIRE = 16, EMERALD = 17, RUBY = 18;

        /// <summary>
        /// These are the block values. They represent the value of each resource
        /// </summary>
        public static int[] BLOCK_VALUES = new int[] { 0, 1, 1, 1, 16, 0, 40, 40, 80, 140, 200, 400, 600, 750, 1000, 1500, 2500, 650, 1300, 1950};

        /// <summary>
        /// SpawnData => MIN_DEPTH, MAX_DEPTH, SPAWN_COUNT, SPAWN_JOIN_RATE
        /// 
        /// These are the spawn datas for each resource in the game
        /// </summary>
        public static SpawnData TIN_SPAWN = new SpawnData(2, 100, 240, 30, TIN);
        public static SpawnData COPPER_SPAWN = new SpawnData(2, 100, 240, 40, COPPER);
        public static SpawnData BAUXITE_SPAWN = new SpawnData(40, 180, 180, 25, BAUXITE);
        public static SpawnData GOLD_SPAWN = new SpawnData(80, 210, 120, 40, GOLD);
        public static SpawnData MITHRIL_SPAWN = new SpawnData(190, 290, 60, 20, MITHRIL);

        public static SpawnData ADAMANTIUM_SPAWN = new SpawnData(200, 290, 50, 40, ADAMANTIUM);
        public static SpawnData COBALT_SPAWN = new SpawnData(220, 290, 70, 35, COBALT);
        public static SpawnData URANIUM_SPAWN = new SpawnData(240, 290, 120, 20, URANIUM);
        public static SpawnData PLUTONIUM_SPAWN = new SpawnData(260, 290, 100, 15, PLUTONIUM);
        public static SpawnData UNOBTANIUM_SPAWN = new SpawnData(280, 290, 30, 10, UNOBTANIUM);

        public static SpawnData SAPPHIRE_SPAWN = new SpawnData(50, 150, 100, 30, SAPPHIRE);
        public static SpawnData EMERALD_SPAWN = new SpawnData(120, 220, 100, 30, EMERALD);
        public static SpawnData RUBY_SPAWN = new SpawnData(180, 280, 100, 30, RUBY);

        public static SpawnData DIRT_SPAWN = new SpawnData(2, -1, 300, 50, DIRT);
        public static SpawnData STONE_SPAWN = new SpawnData(2, 40, 120, 60, STONE, DIRT);
        public static SpawnData GRANITE_SPAWN = new SpawnData(170, 300, 120, 50, STONE);
        public static SpawnData AIR_SPAWN = new SpawnData(4, 300, 200, 50, STONE);

        public static SpawnData COPPER_SPAWN_DIRT = new SpawnData(2, 100, 240, 40, COPPER, DIRT);
        public static SpawnData TIN_SPAWN_DIRT = new SpawnData(2, 100, 240, 30, TIN, DIRT);
        

        /// <summary>
        /// This is used to offset the colour of the gradient, by alpha washing it
        /// </summary>
        public static Color TRANSLUCENT = new Color(Color.White, 0.8f);

        /// <summary>
        /// The coordinate hash of the current block (Active Block only)
        /// </summary>
        public int coordinateHash;

        /// <summary>
        /// The current block's block type (Active Block only)
        /// </summary>
        public int blockType;

        /// <summary>
        /// The current block's maximum and current hitpoint (Active Block only)
        /// </summary>
        protected int maxHP, hp;

        /// <summary>
        /// The value of the current block (Active Block only)
        /// </summary>
        protected int value;

        /// <summary>
        /// These are the timing variables for damaging a block
        /// </summary>
        double healDelay = 0.3;
        double healTimer = 0;

        /// <summary>
        /// This is the block constructor. It is only used to create an Active Block
        /// </summary>
        /// <param name="coordinateHash">The coordinate hash of the block's location</param>
        /// <param name="blockType">The block type to emulate</param>
        public Block(int coordinateHash, int blockType)
        {
            this.blockType = blockType;
            this.coordinateHash = coordinateHash;
            this.collidable = true;
            this.value = BLOCK_VALUES[blockType];
            this.maxHP = GetBlockHP(blockType);
            this.hp = this.maxHP;
        }

        public static void InitaliseBlocks()
        {
            AIR_SPAWN.SetDirectionalWeightings(50, 60, 30, 40);
            DIRT_SPAWN.SetDirectionalWeightings(55, 20, 5, 65);
            STONE_SPAWN.SetDirectionalWeightings(20, 55, 5, 65);
            GRANITE_SPAWN.SetDirectionalWeightings(25, 30, 40, 4);
            TIN_SPAWN_DIRT.SetDirectionalWeightings(25, 25, 7, 45);
            COPPER_SPAWN_DIRT.SetDirectionalWeightings(25, 25, 7, 45);
            GOLD_SPAWN.SetDirectionalWeightings(40, 40, 8, 12);
            COBALT_SPAWN.SetDirectionalWeightings(45, 45, 20, 8);
        }

        public static int GetValue(int id)
        {
            return BLOCK_VALUES[id];
        }

        /// <summary>
        /// This method is used to get the hitpoints of a new block
        /// </summary>
        /// <param name="type">The block type to get the hitpoints of</param>
        /// <returns>The hitpoints of the block</returns>
        private int GetBlockHP(int type)
        {
            switch (type)
            {
                case STONE:
                    return 15;
                case TIN:
                case COPPER:
                case BAUXITE:
                case SAPPHIRE:
                case EMERALD:
                case RUBY:
                    return 25;
                case GOLD:
                case MITHRIL:
                    return 50;
                case ADAMANTIUM:
                case COBALT:
                case URANIUM:
                case PLUTONIUM:
                    return 70;
                case UNOBTANIUM:
                    return 100;
                case AIR:
                case GRASS:
                case DIRT:
                default:
                    return 10;
            }
        }

        /// <summary>
        /// This method is used to check if a block has been destroyed
        /// </summary>
        /// <returns>The boolean representing wether the block has been destroyed</returns>
        public bool Destroyed()
        {
            return hp < 0;
        }

        /// <summary>
        /// This method is used to check if a block has healed
        /// </summary>
        /// <returns>The boolean representing wether the block has been healed</returns>
        public bool Healed()
        {
            return hp >= maxHP;
        }

        /// <summary>
        /// This method is used to damage a block
        /// </summary>
        /// <param name="damage">The amount of damage to deal</param>
        public void Damage(int damage)
        {
            hp -= damage;
            healTimer = healDelay;
        }

        /// <summary>
        /// This method is used to update an Active Block
        /// </summary>
        /// <param name="dt">The delta time since last update</param>
        public override void Update(GameTime dt)
        {
            if (hp < maxHP)
            {
                //If the heal timer has finished, heal
                if (healTimer <= 0)
                {
                    healTimer = healDelay;
                    hp += (maxHP / 10);
                }
            }

            //Run the heal timer
            if (healTimer > 0)
            {
                healTimer -= dt.ElapsedGameTime.TotalSeconds;
            }
        }

        /// <summary>
        /// This method is not used. It is here to satisfy overloading GameObject. It will result in a game crash if attempted to be used
        /// </summary>
        public override void Render(SpriteBatch batch)
        {
            throw new NotSupportedException("Wrong render method.");
        }

        /// <summary>
        /// This is the static render method for Block. It is used to render all blocks and their varients.
        /// </summary>
        /// <param name="batch">The sprite batch to draw to</param>
        /// <param name="x">The x coordinate of the block</param>
        /// <param name="depth">The depth (y coordinate) of the block</param>
        /// <param name="height">The height of the world (For Gradient purposes)</param>
        /// <param name="blockID">The type of block to render</param>
        public static void Render(SpriteBatch batch, int x, int depth, int height, int blockID)
        {

            //This switch statement selects the different type of block and renderes accordingly
            switch (blockID)
            {
                case DIRT:
                    batch.Draw(SpriteManager.tileSet[0, 3], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    break;
                case TIN:
                case COPPER:
                case BAUXITE:
                case GOLD:
                case MITHRIL:
                case ADAMANTIUM:
                case COBALT:
                case URANIUM:
                case PLUTONIUM:
                case UNOBTANIUM:
                    //Draw Stone first, and then the ore overlay
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 0 : 1, 4], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 4 : 5, depth % 2 == 0 ? 0 : 1], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), GetOreColour(blockID));
                    break;
                case SAPPHIRE:
                case EMERALD:
                case RUBY:
                    //Draw Stone first, and then the Gem overlay
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 0 : 1, 4], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    batch.Draw(SpriteManager.tileSet[6, 0], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), GetGemColour(blockID));
                    break;
                case BEDROCK:
                    //Draw Stone first, and then the Bedrock Overlay
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 0 : 1, 4], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    batch.Draw(SpriteManager.tileSet[(x + depth * Math.Abs(blockID - x)) % 3, 0], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    break;
                case GRANITE:
                    batch.Draw(SpriteManager.tileSet[3, 3], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.Gray);
                    break;
                case STONE:
                default:
                    //This is the default case. it will just render blank stone
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 0 : 1, 4], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    break;
            }

            //Mostly for debug purposes, this is used to check if the gradient is to be drawn            
            if (DRAW_GRADIENT)
            {
                //The Gradient texture is 300px tall. So Scale accordingly
                if (height != 300)
                {
                    depth /= height / 300;
                }

                //Draw the gradient on top of the tile just drawn, selecting the 1x1px square from the gradient texture at x - 0, y - depth
                batch.Draw(SpriteManager.gradient, new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), new Rectangle(0, depth, 1, 1), TRANSLUCENT);
            }

            //This switch statement selects the different type of block and renderes accordingly
            switch (blockID)
            {
                case TIN:
                case COPPER:
                case BAUXITE:
                case GOLD:
                case MITHRIL:
                case ADAMANTIUM:
                case COBALT:
                case URANIUM:
                case PLUTONIUM:
                case UNOBTANIUM:
                    batch.Draw(SpriteManager.tileSet[x % 2 == 0 ? 4 : 5, depth % 2 == 0 ? 0 : 1], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), GetOreColour(blockID));
                    break;
                case SAPPHIRE:
                case EMERALD:
                case RUBY:
                    batch.Draw(SpriteManager.tileSet[6, 0], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), GetGemColour(blockID));
                    break;
                case BEDROCK:
                    batch.Draw(SpriteManager.tileSet[(x + depth * Math.Abs(blockID - x)) % 3, 0], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                    break;
                case GRANITE:
                    break;
                case STONE:
                default:
                    break;
            }

        }

        /// <summary>
        /// This method is used to render an Active Block
        /// </summary>
        /// <param name="batch">The batch to render to</param>
        /// <param name="x">The x coordinate of the block</param>
        /// <param name="depth">The depth of the block</param>
        /// <param name="height">The height of the block</param>
        public void RenderActive(SpriteBatch batch, int x, int depth, int height)
        {
            Block.Render(batch, x, depth, height, blockType);

            if (hp < maxHP)
            {
                int frame = MathHelper.Clamp(SpriteManager.destroyTileAnim.Length - (hp / (maxHP / SpriteManager.destroyTileAnim.Length)), 0, SpriteManager.destroyTileAnim.Length - 1);
                //batch.Draw(SpriteManager.uiBG, new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.White);
                batch.Draw(SpriteManager.destroyTileAnim[frame], new Rectangle(x * GameScene.TILESIZE, depth * GameScene.TILESIZE, GameScene.TILESIZE, GameScene.TILESIZE), Color.Gray);
            }
        }

        /// <summary>
        /// This method is used to grab the ore tint colour
        /// </summary>
        /// <param name="tileID">The tile id to grab the colour of</param>
        /// <returns>The colour of the ore</returns>
        public static Color GetOreColour(int tileID)
        {
            switch (tileID)
            {
                case TIN:
                    return Color.Gray;
                case COPPER:
                    return Color.Orange;
                case BAUXITE:
                    return Color.Silver;
                case GOLD:
                    return Color.Gold;
                case MITHRIL:
                    return Color.Blue;
                case ADAMANTIUM:
                    return Color.DarkGreen;
                case COBALT:
                    return Color.Cyan;
                case URANIUM:
                    return Color.LimeGreen;
                case PLUTONIUM:
                    return Color.DeepPink;
                case UNOBTANIUM:
                    return Color.Yellow;
            }

            return Color.White;
        }

        /// <summary>
        /// This method is used to grab the gem tint colour
        /// </summary>
        /// <param name="tileID">The tile id to grab the colour of</param>
        /// <returns>The colour of the gem</returns>
        public static Color GetGemColour(int tileID)
        {
            switch (tileID)
            {
                case SAPPHIRE:
                    return Color.Blue;
                case EMERALD:
                    return Color.Green;
                case RUBY:
                    return Color.Red;
            }

            return Color.White;
        }
    }
}
