﻿using GGJ_2015.GameObjects.Entities;
using GGJ_2015.GameObjects.Entities.Projectiles;
using GGJ_2015.Scenes;
using GGJ_2015.UI.CustomElements;
using GGJ_2015.Util;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015
{
    public class Player : Entity, Camera2D.ITargetable
    {
        private Vector2 origin = new Vector2(0, 0);
        private const int LEFT = 0, RIGHT = 1;

        /// <summary>
        /// Score/Money tbd
        /// </summary>
        public float score = 0;

        private int facing = LEFT;
        private int damage = 5;
        //The Max values must be devisible by 30
        public int hp = 25, maxHP = 30;
        public int fuel = 120, maxFuel = 120;
        int fuelDist = 100;
        protected double travelMod = 0;
        public int inv = 0, maxInv = 90;

        protected int range = 5;
        protected double laserRotation = 0;
        protected Vector2 mouseScreenPos = Vector2.Zero;
        protected Vector2 toMouse = Vector2.Zero;

        protected double interactTimer = 0;
        protected double interactDelay = 0.2;

        private Upgrades upgrades = new Upgrades();
        protected GameScene scene;
        protected Inventory inventory;

        private Animation playerAnim;
        private List<Laser> lasers = new List<Laser>();

        public bool inShop = true;
        private ShopUI shop;

        /// <summary>
        /// This is the constructor for the player
        /// </summary>
        /// <param name="inPosition">The positon to create the player at</param>
        /// <param name="inScene">The GameScene the player will be in</param>
        public Player(Vector2 inPosition, GameScene inScene)
        {
            this.position = inPosition;
            this.scene = inScene;
            this.acceleration = new Vector2(5, 10);
            this.size = new Vector2(96, 64);

            inventory = new Inventory(upgrades.InventorySize);
            playerAnim = new Animation(SpriteManager.player, 0.15);

            shop = new ShopUI(this, new Vector2(600, 400));
        }

        /// <summary>
        /// This method is used to get the current damage the player can deal
        /// </summary>
        /// <returns>The amount of damage the player can deal (in one hit)</returns>
        public int GetDamage()
        {
            return damage;
        }

        /// <summary>
        /// This method updates the player
        /// </summary>
        /// <param name="dt">The delta time since the last update</param>
        public override void Update(GameTime dt)
        {
            //Run the interaction timer
            if (interactTimer > 0)
            {
                interactTimer -= dt.ElapsedGameTime.TotalSeconds;
            }
            if (inShop)
            {
                shop.Update(dt);
            }
            else
            {

                if (InputManager.e.WasReleased() && position.Y < 100) inShop = true;
                if (InputManager.o.WasReleased() && position.Y < 100)
                {
                    score += inventory.GetValue();
                    inventory.ClearInventory();
                }
                #region Keyboard Handling
                //If Left/A or Right/D are down, accelerate in the relevant direction
                if (InputManager.right.IsDown() || InputManager.d.IsDown())
                {
                    facing = RIGHT;
                    velocity.X += (float)(acceleration.X * dt.ElapsedGameTime.TotalSeconds);
                }
                else if (InputManager.left.IsDown() || InputManager.a.IsDown())
                {
                    facing = LEFT;
                    velocity.X -= (float)(acceleration.X * dt.ElapsedGameTime.TotalSeconds);
                }
                else
                {
                    velocity.X -= (float)(velocity.X * dt.ElapsedGameTime.TotalSeconds);
                    if (Math.Abs(velocity.X) < 0.1)
                    {
                        velocity.X = 0;
                    }
                }

                //If Up/W is down, try to hover/fly
                if (InputManager.up.IsDown() || InputManager.w.IsDown())
                {
                    velocity.Y -= (float)(acceleration.Y * dt.ElapsedGameTime.TotalSeconds);
                }
                else
                {
                    velocity.Y += (float)(scene.GRAVITY * dt.ElapsedGameTime.TotalSeconds);
                }

                //Test collision in the x axis
                if (!scene.TestCollision(position.X + velocity.X, position.Y, size))
                {
                    position.X += velocity.X;
                    travelMod += Math.Abs(velocity.X);
                }
                else
                {
                    //Collision has occured, so set x velocity to 0, and rebound 2 pixels in the opposite direction
                    //The rebound is to stop the player getting stuck inside a block
                    position.X += 2 * velocity.X < 0 ? 1 : -1;
                    velocity.X = 0;
                }

                //Test collision in the y axis
                if (!scene.TestCollision(position.X, position.Y + velocity.Y, size))
                {
                    position.Y += velocity.Y;
                    if (velocity.Y < 0)
                    {
                        travelMod += Math.Abs(velocity.Y);
                    }
                }
                else
                {
                    //Collision has occured, so set y velocity to 0
                    velocity.Y = 0;
                }

                if (travelMod > 0)
                {
                    while (travelMod > fuelDist)
                    {
                        travelMod -= fuelDist;
                        fuel--;

                        if (fuel < 0)
                        {
                            GameOver();
                        }
                    }
                }

                #endregion

                #region Mouse Management
                //Perform vector maths on the mouse in order to accertain its world coordinates and the angle of rotation
                mouseScreenPos = InputManager.mousePos;
                toMouse = new Vector2(mouseScreenPos.X - (GGJGame.GWIDTH / 2), mouseScreenPos.Y - (GGJGame.GHEIGHT / 2));
                laserRotation = Math.Atan2(-toMouse.Y, toMouse.X) + (Math.PI / 2);

                //If left mouse button is down, interact with a block
                if (InputManager.lmb.IsDown())
                {
                    //Check to see if the interaction timer has expired
                    if (interactTimer <= 0)
                    {
                        //Adjust mouse coordinate to actual world position
                        Vector2 adjustedMouse = new Vector2((position.X + toMouse.X + size.X / 2), (position.Y + toMouse.Y + size.Y / 2));

                        //Attempt to interact with a block
                        //Pass the coordinates of the CENTER of the player
                        if (scene.InteractWithBlock(adjustedMouse, Position, range))
                        {
                            toMouse.Normalize();
                            lasers.Add(new Laser(Position, adjustedMouse, toMouse));
                        }


                        //Reset the interaction timer
                        interactTimer = interactDelay;
                    }
                }
                #endregion

                playerAnim.Update(dt);

                for (int i = 0; i < lasers.Count; i++)
                {
                    lasers[i].Update(dt);
                    if (lasers[i].remove)
                    {
                        lasers.RemoveAt(i);
                    }
                }

                //Ensure the velocity is within terminal bounds
                ClampVelocity();

                //Wrap the x coordinate around the world
                if (position.X < 0) position.X += scene.GetWorldWidth();
                if (position.X > scene.GetWorldWidth()) position.X -= scene.GetWorldWidth();
            }
        }

        /// <summary>
        /// This is the player render method
        /// </summary>
        /// <param name="batch">The batch to draw the player to.</param>
        public override void Render(SpriteBatch batch)
        {
            batch.Draw(playerAnim.GetCurrentFrame(), new Rectangle((int)position.X, (int)position.Y, (int)size.X, (int)size.Y), null, Color.White, 0, origin, facing == RIGHT ? SpriteEffects.FlipHorizontally : SpriteEffects.None, 0);

            for (int i = 0; i < lasers.Count; i++)
            {
                lasers[i].Render(batch);
            }

            //Draw the mouse pointer
            batch.Draw(SpriteManager.debugTexture, new Rectangle((int)(position.X + toMouse.X + size.X / 2 - 4), (int)(position.Y + toMouse.Y + size.Y / 2 - 4), 8, 8), Color.Red);
        }

        public void RenderUI(SpriteBatch batch)
        {
            BitmapFont.font.Draw(batch, 0, 0, "Money: " + GetMoney(), Color.Red);
            if (inShop) shop.Render(batch);
        }

        /// <summary>
        /// This is the implementation of the ITargetable interface. It returns a vector representing the center of the player
        /// </summary>
        public Vector2 Position
        {
            get { return new Vector2(position.X + size.X / 2, position.Y + size.Y / 2); }
        }

        /// <summary>
        /// This method is used to check the rotation of the player for the camera. Not used.
        /// </summary>
        public float Rotation
        {
            get { return 0; }
        }

        public int ItemsInInventory()
        {
            return inventory.GetInventoryCount();
        }

        public int InventorySize()
        {
            return inventory.MaxSize();
        }

        public void AddToInventory(int blockID, int quantity)
        {
            inventory.AddItem(blockID, quantity);
        }

        public Upgrades GetUpgrades()
        {
            return upgrades;
        }

        public int GetMoney()
        {
            return (int)score;
        }

        public void FillFuel()
        {
            fuel = maxFuel;
        }

        public void UpgradeInventory()
        {
            score -= upgrades.upgradeCosts[upgrades.inventoryTier];
            upgrades.inventoryTier++;
            inventory.inventorySize = upgrades.InventorySize;
        }

        public void UpgradeEngine()
        {
            score -= upgrades.upgradeCosts[upgrades.engineTier];
            upgrades.engineTier++;
            acceleration = upgrades.engineSpeed[upgrades.engineTier];
        }
        public void UpgradeLaser()
        {
            score -= upgrades.upgradeCosts[upgrades.laserTier];
            upgrades.laserTier++;
            damage = upgrades.laserDamage[upgrades.laserTier];
        }
        public void UpgradeShield()
        {
            score -= upgrades.upgradeCosts[upgrades.shieldTier];
            upgrades.shieldTier++;
            hp = upgrades.shieldUpgrades[upgrades.shieldTier];
        }

        public void UpgradeFuel()
        {
            score -= upgrades.upgradeCosts[upgrades.fuelTier];
            upgrades.fuelTier++;
            maxFuel = upgrades.fuelSize[upgrades.fuelTier];
        }



        public void GameOver() { }
    }
}
