﻿using GGJ_2015.Environment;
using GGJ_2015.UI;
using GGJ_2015.UI.CustomElements;
using GGJ_2015.Util;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015.Scenes
{
    public class GameScene : Scene
    {
        public const bool DEBUG = false;
        public static int TILESIZE = 32;
        public float GRAVITY = 9.8f;
        public int width, height = 300;
        protected int[,] tiles;

        protected Camera2D cam;
        protected Player player;

        protected List<Block> activeBlocks = new List<Block>();
        
        #region UI
        public UIFrame barFrame;
        public UIElement hpBar;
        public UIElement fuelBar;
        public UIElement invBar;
        #endregion


        public GameScene(int w, Camera2D cam)
        {
            this.cam = cam;

            player = new Player(new Vector2(0, -100), this);
            cam.Target = player;
            GenerateUI();
            GenerateMap(w, DateTime.Now.Ticks);
        }

        public void GenerateUI()
        {
            uiElements = new List<UIElement>();

            barFrame = new UIFrame(new Vector2(20, 20));
            barFrame.AddElement(new HPBar(player), new Vector2(20, 20));
            barFrame.AddElement(new FuelBar(player), new Vector2(20, 50));
            barFrame.AddElement(new InvBar(player), new Vector2(20, 80));



            uiElements.Add(barFrame);
        }

        #region Map Generation
        /// <summary>
        /// This method is used to generate a map
        /// </summary>
        /// <param name="w">The width of the map</param>
        /// <param name="h">The height of the map</param>
        private void GenerateMap(int w, long seed)
        {
            this.width = w;

            tiles = new int[width, height];

            Random rand = new Random(seed.GetHashCode());

            WorldGenerator generator = new WorldGenerator(300, 300);
            tiles = generator.Generate();

            //For debugging, create a tunnel through the entire map to check ore spawns etc
            if (DEBUG)
            {
                int xStart = 5;
                for (int y = 0; y < height - 1; y++)
                {
                    for (int x = 0; x < 8; x++)
                    {
                        tiles[Normalise(x + xStart, 0, width), y] = -1;
                    }

                    xStart++;
                }
            }
        }

        #endregion


        /// <summary>
        /// This method is used to get the active block at a coordinate
        /// </summary>
        /// <param name="x">The x coordinate</param>
        /// <param name="y">The y coordinate</param>
        /// <returns>The active block (if any) at the coordinate</returns>
        private Block BlockAt(int x, int y)
        {
            //Hash the passed coordinates
            int hash = CoordinateHash(x, y);

            //Loop through the active block list
            for (int i = 0; i < activeBlocks.Count; i++)
            {
                //Check each active block against the coordinate hash
                if (activeBlocks[i].coordinateHash == hash)
                {
                    //If matched, return it
                    return activeBlocks[i];
                }
            }

            //No match so return null to show this
            return null;
        }

        /// <summary>
        /// This method is used to interact with an individual block
        /// </summary>
        /// <param name="pos">The location to interact with</param>
        /// <param name="playerPos">This is the current location of the center of the player</param>
        /// <param name="playerRange">This is the range of the player in blocks</param>
        public bool InteractWithBlock(Vector2 pos, Vector2 playerPos, int playerRange)
        {
            //Convert the passed position to block coordinates and normalise the x
            int x = Normalise((int)(pos.X / 32), 0, width);
            int y = (int)(pos.Y / 32);

            //Normalise the player xPositon and range check the y
            int playerX = Normalise((int)(playerPos.X / 32), 0, width);
            int playerY = (int)(playerPos.Y / 32);

            //If the block is out of range, return false as no interaction occured.
            if (!RangeCheck(x, y, playerX, playerY, playerRange)) return false;

            //Range Check the y coordinate
            if (y < 0 || y >= height) return false;

            //Check to see if there is an active block at that coordinate
            Block b = BlockAt(x, y);

            if (b == null)
            {
                //If the tile is air, return false as no interaction
                if (tiles[x, y] == -1) return false;

                //Create a new active block at the location
                b = new Block(CoordinateHash(x, y), tiles[x, y]);

                //Add the new block the active block list
                activeBlocks.Add(b);

                //Damage the new block
                b.Damage(player.GetDamage());
            }
            else
            {
                //Active Block found, damage it
                b.Damage(player.GetDamage());
            }

            return true;
        }

        /// <summary>
        /// This method is used to smooth over range checking with the interaction of the player with the world.
        /// It is required because of the edge case with the world wrapping back to itself, the deltaX can be 
        /// quite large, when it isnt meant to be. This is solved by some normalisation maths.
        /// </summary>
        /// <param name="x">The x of the target block</param>
        /// <param name="y">The y of the target block</param>
        /// <param name="playerX">The player x</param>
        /// <param name="playerY">The player y</param>
        /// <param name="playerRange">The player's range</param>
        /// <returns>A boolean representing wether or not the block is in range</returns>
        public bool RangeCheck(int x, int y, int playerX, int playerY, int playerRange)
        {
            int dx = 0, dx2 = 0, dy = 0;

            int nX = Normalise(x, 0, width);
            dx = nX - playerX;
            dx2 = (playerX + (playerX > nX ? -1 : 1) * width) - nX;
            dy = y - playerY;

            return (Math.Abs(dx) < playerRange || Math.Abs(dx2) < playerRange) && Math.Abs(dy) < playerRange;
        }

        /// <summary>
        /// This method is used to update the game scene
        /// </summary>
        /// <param name="dt">The delta time since last update</param>
        public override void Update(GameTime dt)
        {
            //Update the player
            player.Update(dt);

            //Loop through and update the active blocks
            for (int i = 0; i < activeBlocks.Count; i++)
            {
                activeBlocks[i].Update(dt);

                //If the block has been destroyed, remove it.
                if (activeBlocks[i].Destroyed())
                {
                    //Update the tile array to show the block is now air (destroyed)
                    SetTileFromHash(activeBlocks[i].coordinateHash, -1);
                    player.AddToInventory(activeBlocks[i].blockType, 1);
                    //Remove the active block
                    activeBlocks.RemoveAt(i);
                }
                else if (activeBlocks[i].Healed())
                {
                    //Check to see if the block has healed, and if it has, remove it
                    activeBlocks.RemoveAt(i);

                    //There is no update to the block array here, as there is no change to the 
                    //type of block at the location
                }
            }
        }

        #region Rendering
        /// <summary>
        /// This method is used to render the map
        /// </summary>
        /// <param name="batch">The current spritebatch</param>
        public override void Render(SpriteBatch batch)
        {
            //Find the start x of the rendering (In terms of block)
            int xStart = (int)(player.Position.X - cam.Width / 2) / 32;
            int yStart = (int)(player.Position.Y - cam.Height / 2) / 32;

            //The render width (in terms of blocks)
            int rw = (int)cam.Width / 32 + 2;
            int rh = (int)cam.Height / 32 + 2;

            //Loop through the x and y values to render
            for (int x = xStart - 1; x < xStart + rw; x++)
            {
                //Normalise the x to get a tile index
                int rx = Normalise(x, 0, width);
                for (int y = yStart - 1; y < yStart + rh; y++)
                {
                    //Range check the y
                    if (y < 0 || y >= height) continue;

                    //Check to see if the relevant tile is air
                    if (GetTile(rx, y) == -1) continue;

                    //Check to see if there is an active block
                    Block b = BlockAt(x, y);
                    if (b != null)
                    {
                        //If there is an active block, render it
                        b.RenderActive(batch, x, y, height);
                    }
                    else
                    {
                        //Run the generic block render method
                        Block.Render(batch, x, y, height, tiles[rx, y]);
                    }
                }
            }

            //Render the player
            player.Render(batch);
        }

        #endregion

        /// <summary>
        /// This method is used to generate a unique hash for each tile, by converting it
        /// to a 1d array index
        /// </summary>
        /// <param name="x">The input x value</param>
        /// <param name="y">The input y value</param>
        /// <returns>The unique hash of the coordinate</returns>
        public int CoordinateHash(int x, int y)
        {
            return x + y * width;
        }

        /// <summary>
        /// This method is used to change a tile in the main tile array. Usually called 
        /// from the update method as the block in question will have the block hash 
        /// instead of the x, y coordinate of the block.
        /// </summary>
        /// <param name="hash">The block hash</param>
        /// <param name="newTile">The tile index to set</param>
        public void SetTileFromHash(int hash, int newTile)
        {
            tiles[hash % width, hash / width] = newTile;
        }

        /// <summary>
        /// This method will Normalise a value (input) between min and max
        /// </summary>
        /// <param name="input">The value to normalise</param>
        /// <param name="min">The minimum value</param>
        /// <param name="max">The maximum value</param>
        /// <returns>The normalized value</returns>
        public int Normalise(int input, int min, int max)
        {
            //while (input >= max) input -= (max - min);
            //while (input < min) input += (max - min);
            input %= max;
            if (input < min) input += max;
            return input;
        }

        /// <summary>
        /// This method will check to see if a tile is collideable
        /// </summary>
        /// <param name="x">The x location to check</param>
        /// <param name="y">The y location to check</param>
        /// <returns>
        ///     The boolean representing wether or not the location is collideable
        /// </returns>
        public bool GetCollideable(int x, int y)
        {
            x = Normalise(x, 0, width);
            if (y < 0 || y >= height) return false;
            return tiles[x, y] != -1;
        }

        /// <summary>
        /// This method is used to auto cast the float parameters to int for the get
        /// collideable method
        /// </summary>
        /// <param name="x">The x location to check</param>
        /// <param name="y">The y location to check</param>
        /// <returns>
        ///     The boolean representing wether or not the location is collideable
        /// </returns>
        public bool GetCollideable(float x, float y)
        {
            return GetCollideable((int)x, (int)y);
        }

        /// <summary>
        /// This method is used to safely get a tile from the tiles array. It is safe as it will validate
        /// the coordinate passed to it by normalising the x value and range checking the y value.
        /// </summary>
        /// <param name="x">The x coordinate to retrieve</param>
        /// <param name="y">The y coordinate to retrieve</param>
        /// <returns>The retrieved block</returns>
        public int GetTile(int x, int y)
        {
            if (x < 0 || x >= width) return -1;
            if (y < 0 || y >= height) return -1;
            return tiles[x, y];
        }

        /// <summary>
        /// This method is used to test collision detection of a specified location with a set size
        /// </summary>
        /// <param name="x">The x coordinate to check from</param>
        /// <param name="y">The y coordinate to check from</param>
        /// <param name="size">The size to check</param>
        /// <returns>A boolean representing wether a collision has occured or not</returns>
        public bool TestCollision(float x, float y, Vector2 size)
        {
            //Make the collision box slightly smaller to make getting into
            //same size locations slightly easier
            x += 2;
            y += 2;
            size.X -= 4;
            size.Y -= 4;

            //Loop through all tiles that could be involved in collision
            for (int x1 = (int)x / TILESIZE; x1 < (x + size.X) / TILESIZE; x1++)
            {
                for (int y1 = (int)y / TILESIZE; y1 < (y + size.Y) / TILESIZE; y1++)
                {
                    //Check to see if the resultant coordinate is hitting a collideable tile
                    if (GetCollideable(x1, y1)) return true;
                }
            }

            //No collision has occured
            return false;
        }

        /// <summary>
        /// This method will return the width of the world in actual terms of TileWidth * TileSize
        /// </summary>
        /// <returns>The width of the world</returns>
        public int GetWorldWidth()
        {
            return width * TILESIZE;
        }

        public override void RenderUI(SpriteBatch batch)
        {
            if (uiElements != null)
            {
                for (int i = 0; i < uiElements.Count; i++)
                {
                    uiElements[i].Render(batch);
                }
            }

            player.RenderUI(batch);
        }
    }
}
