﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015.Util
{
    public class Camera2D
    {
        public Camera2D(Viewport inViewport)
        {
            this.viewport = inViewport;
            this.origin = new Vector2(viewport.Width / 2.0f, viewport.Height / 2.0f);
        }

        /// <summary>
        /// Gets or sets the position of the camera.
        /// </summary>
        public Vector2 Position
        {
            get
            {
                return position;
            }
            set
            {
                position = value;
                ValidatePosition();
            }
        }

        /// <summary>
        /// This is the getter for the camera width based on the viewport
        /// </summary>
        public float Width
        {
            get
            {
                return viewport.Width;
            }
        }

        /// <summary>
        /// This is the getter for the camera height based on the viewport
        /// </summary>
        public float Height
        {
            get { return viewport.Height; }
        }

        /// <summary>
        /// Gets or sets the zoom of the camera.
        /// </summary>
        public float Zoom
        {
            get
            {
                return zoom;
            }
            set
            {
                zoom = MathHelper.Max(value, MinZoom);
                ValidateZoom();
                ValidatePosition();
            }
        }

        /// <summary>
        /// Sets a rectangle that describes which region of the world the camera should
        /// be able to see. Setting it to null removes the limit.
        /// </summary>
        public Rectangle? Limits
        {
            set
            {
                limits = value;
                ValidateZoom();
                ValidatePosition();
            }
        }

        /// <summary>
        /// Calculates a view matrix for this camera.
        /// </summary>
        public Matrix ViewMatrix
        {
            get
            {
                return Matrix.CreateTranslation(new Vector3(-position, 0f)) *
                       Matrix.CreateTranslation(new Vector3(-origin, 0f)) *
                       Matrix.CreateScale(zoom, zoom, 1f) *
                       Matrix.CreateTranslation(new Vector3(origin, 0f));
            }
        }

        /// <summary>
        /// This is a getter for the TargetPos of the camera
        /// </summary>
        public Vector2 TargetPos
        {
            get
            {
                if (target == null) return new Vector2(0, 0);
                Vector2 targetProjection = target.Position;
                return new Vector2(targetProjection.X - origin.X, targetProjection.Y - origin.Y);
            }
        }

        public ITargetable Target
        {
            get { return target; }
            set { target = value; }
        }

        /// <summary>
        /// When using limiting, makes sure the camera position is valid.
        /// </summary>
        private void ValidatePosition()
        {
            if (limits.HasValue)
            {
                Vector2 cameraWorldMin = Vector2.Transform(Vector2.Zero, Matrix.Invert(ViewMatrix));
                Vector2 cameraSize = new Vector2(viewport.Width, viewport.Height) / zoom;
                Vector2 limitWorldMin = new Vector2(limits.Value.Left, limits.Value.Top);
                Vector2 limitWorldMax = new Vector2(limits.Value.Right, limits.Value.Bottom);
                Vector2 positionOffset = position - cameraWorldMin;
                position = Vector2.Clamp(cameraWorldMin, limitWorldMin, limitWorldMax - cameraSize) + positionOffset;
            }
        }

        /// <summary>
        /// When using limiting, makes sure the camera zoom is valid.
        /// </summary>
        private void ValidateZoom()
        {
            if (limits.HasValue)
            {
                float minZoomX = (float)viewport.Width / limits.Value.Width;
                float minZoomY = (float)viewport.Height / limits.Value.Height;
                zoom = MathHelper.Max(zoom, MathHelper.Max(minZoomX, minZoomY));
            }
        }

        /// <summary>
        /// This method is used to make the camera look at a position
        /// </summary>
        /// <param name="position">The position vector to look at</param>
        public void LookAt(Vector2 position)
        {
            Position = position - origin;
        }

        /// <summary>
        /// This method will get the view rectangle of the current camera
        /// </summary>
        /// <param name="mod">This is a modifier for the size of the rectangle</param>
        /// <returns>A rectangle representing what is visible</returns>
        public Rectangle GetViewPort(int mod = 0)
        {
            return new Rectangle((int)position.X - mod, (int)position.Y - mod, viewport.Width * (int)(1 / zoom) + mod * 2, viewport.Height * (int)(1 / zoom) + mod * 2);
        }

        /// <summary>
        /// This method updates the camera, it is used for the easing of the camera, when following a target
        /// </summary>
        /// <param name="dt">The delta time since last update</param>
        public void Update(GameTime dt)
        {
            if (target != null)
            {
                if (tweenToTarget)
                {
                    position.X += (TargetPos.X - position.X) * moveSpeed * (float)dt.ElapsedGameTime.TotalSeconds;
                    position.Y += (TargetPos.Y - position.Y) * moveSpeed * (float)dt.ElapsedGameTime.TotalSeconds;
                    ValidatePosition();
                }
                else
                {
                    position = TargetPos;
                }
            }
        }

        public interface ITargetable
        {
            Vector2 Position { get; }

            float Rotation { get; }
        }


        private const float MinZoom = 0.01f;

        private readonly Viewport viewport;
        private readonly Vector2 origin;

        private Vector2 position;
        private float zoom = 1f;
        private Rectangle? limits;
        private ITargetable target;
        private float moveSpeed = 0.5f;
        private bool tweenToTarget = false;
    }

}
