﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace GGJ_2015.Util
{
    /// <summary>
    /// This class is used to be a static store of texture references to save loading them multiple times.
    /// </summary>
    public class SpriteManager
    {
        public static Texture2D[] font;

        public static Texture2D debugTexture;
        public static Texture2D dirtTexture;

        public static Texture2D uiBG;
        public static Texture2D hudUpgrades;

        public static Texture2D[,] tileSet;
        public static Texture2D[] destroyTileAnim;
        
        public static Texture2D hpBarL, hpBarR, hpBarM;
        public static Texture2D fuelBarL, fuelBarR, fuelBarM;
        public static Texture2D invBarL, invBarR, invBarM;

        public static Texture2D[] miningLaser;
        public static Texture2D[] player;

        public static Texture2D background;
        public static Texture2D gradient;

        public static Texture2D shopSkeleton;

        public static Texture2D[] buyButton;

        public static Texture2D[] laserButton;
        public static Texture2D[] shieldButton;
        public static Texture2D[] engineButton;
        public static Texture2D[] inventoryButton;
        public static Texture2D[] fuelButton;

        public static Texture2D buyFuelButton;
        public static Texture2D exitButton;

        public static void Initialise(ContentManager content, GraphicsDevice graphicsDevice)
        {
            debugTexture = new Texture2D(graphicsDevice, 1, 1);
            debugTexture.SetData<Color>(new[] { Color.White });

            dirtTexture = content.Load<Texture2D>("Sprites/groundBaseGrey");
            font = Cut2DTilesTo1D("Sprites/Font", 26, 2, content, graphicsDevice);
            BitmapFont.font = new BitmapFont(font, "abcdefghijklmnopqrstuvwxyz0123456789.:!", Color.White);

            hudUpgrades = Load("Sprites/hud_upgrades_final", content);
            uiBG = Load("Sprites/hudBackground", content);

            hpBarL = Load("Sprites/UI/barHorizontal_red_left", content);
            hpBarM = Load("Sprites/UI/barHorizontal_red_mid", content);
            hpBarR = Load("Sprites/UI/barHorizontal_red_right", content);

            fuelBarL = Load("Sprites/UI/barHorizontal_blue_left", content);
            fuelBarM = Load("Sprites/UI/barHorizontal_blue_mid", content);
            fuelBarR = Load("Sprites/UI/barHorizontal_blue_right", content);

            invBarL = Load("Sprites/UI/barHorizontal_green_left", content);
            invBarM = Load("Sprites/UI/barHorizontal_green_mid", content);
            invBarR = Load("Sprites/UI/barHorizontal_green_right", content);

            tileSet = CutTiles("Sprites/tileSheet", 10, 10, content, graphicsDevice);
            destroyTileAnim = CutTiles("Sprites/tileDestroyAnim", 7, content, graphicsDevice);

            miningLaser = CutTiles("Sprites/finishedLaser", 8, content, graphicsDevice);
            player = CutTiles("Sprites/playerPrelim", 3, content, graphicsDevice);

            gradient = Load("Sprites/gradient", content);
            background = Load("Sprites/background1", content);

            shopSkeleton = Load("Sprites/UI/Shop/shopSkeleton", content);
            exitButton = Load("Sprites/UI/Shop/btnExit", content);
            buyFuelButton = Load("Sprites/UI/Shop/btnBuyFuel", content);

            buyButton = CutTiles("Sprites/UI/Shop/buyButton", 2, content, graphicsDevice);

            shieldButton = CutTiles("Sprites/UI/Shop/btnShield", 2, content, graphicsDevice);
            laserButton = CutTiles("Sprites/UI/Shop/btnLaser", 2, content, graphicsDevice);
            engineButton = CutTiles("Sprites/UI/Shop/btnEngine", 2, content, graphicsDevice);
            inventoryButton = CutTiles("Sprites/UI/Shop/btnInventory", 2, content, graphicsDevice);
            fuelButton = CutTiles("Sprites/UI/Shop/btnFuel", 2, content, graphicsDevice);
        }

        /// <summary>
        /// This method will load a Texture from the passed file and content manager. This method is 
        /// used to allow for ease of validation of the loaded texture if needed.
        /// </summary>
        /// <param name="file">The texture name</param>
        /// <param name="content">The content manager to load with</param>
        /// <returns>The loaded texture</returns>
        public static Texture2D Load(String file, ContentManager content)
        {
            return content.Load<Texture2D>(file);
        }

        /// <summary>
        /// This method will load a 2D SpriteSheet and convert it to a 1D array of textures
        /// </summary>
        /// <param name="texture">The SpriteSheet to load</param>
        /// <param name="xTiles">The number of tiles in the x axis</param>
        /// <param name="yTiles">The number of tiles in the y axis</param>
        /// <param name="content">The content manager to load from</param>
        /// <param name="graphics">The graphics device (GPU) of the current pc. Used to generate new textures</param>
        /// <returns>A 1D array of textures representing a 2D SpriteSheet</returns>
        public static Texture2D[] Cut2DTilesTo1D(String texture, int xTiles, int yTiles, ContentManager content, GraphicsDevice graphics)
        {
            //Create an empty 1D array
            Texture2D[] result = new Texture2D[xTiles * yTiles];

            //Load the 2D array of textures
            Texture2D[,] temp = CutTiles(texture, xTiles, yTiles, content, graphics);

            //Loop through the 2D array
            for (int x = 0; x < xTiles; x++)
            {
                for (int y = 0; y < yTiles; y++)
                {
                    //And set the 1D array accordingly
                    result[x + y * xTiles] = temp[x, y];
                }
            }
            
            //Return the 1D array
            return result;
        }

        /// <summary>
        /// This method will cut a strip of tiles into seperate textures
        /// </summary>
        /// <param name="texture">The original texture</param>
        /// <param name="xTiles">The width in tiles</param>
        /// <param name="content">The current content manager</param>
        /// <param name="graphicsDevice">The graphics manager to use</param>
        /// <returns>An array representing the spritesheet</returns>
        public static Texture2D[] CutTiles(String texture, int xTiles, ContentManager content, GraphicsDevice graphicsDevice)
        {
            Texture2D[] result = new Texture2D[xTiles];
            Texture2D originalTexture = content.Load<Texture2D>(texture);
            int width = originalTexture.Width / xTiles;

            for (int x = 0; x < xTiles; x++)
            {

                Rectangle sourceRectangle = new Rectangle(x * width, 0, width, originalTexture.Height);

                Texture2D cropTexture = new Texture2D(graphicsDevice, sourceRectangle.Width, sourceRectangle.Height);
                Color[] data = new Color[sourceRectangle.Width * sourceRectangle.Height];
                originalTexture.GetData(0, sourceRectangle, data, 0, data.Length);
                cropTexture.SetData(data);

                result[x] = cropTexture;

            }

            originalTexture.Dispose();

            return result;
        }

        /// <summary>
        /// This method will cut out individual tiles from a single texture
        /// </summary>
        /// <param name="texture">The original texture</param>
        /// <param name="xTiles">The width in tiles</param>
        /// <param name="yTiles">The height in tiles</param>
        /// <param name="content">The content manager to load from</param>
        /// <param name="graphicsDevice">The current graphics device</param>
        /// <returns>A 2D array of textures representing the spritesheet</returns>
        public static Texture2D[,] CutTiles(String texture, int xTiles, int yTiles, ContentManager content, GraphicsDevice graphicsDevice)
        {
            Texture2D[,] result = new Texture2D[xTiles, yTiles];
            Texture2D originalTexture = content.Load<Texture2D>(texture);

            int width = originalTexture.Width / xTiles;
            int height = originalTexture.Height / yTiles;

            for (int x = 0; x < xTiles; x++)
            {
                for (int y = 0; y < yTiles; y++)
                {
                    Rectangle sourceRectangle = new Rectangle(x * width, y * height, width, height);

                    Texture2D cropTexture = new Texture2D(graphicsDevice, sourceRectangle.Width, sourceRectangle.Height);
                    Color[] data = new Color[sourceRectangle.Width * sourceRectangle.Height];
                    originalTexture.GetData(0, sourceRectangle, data, 0, data.Length);
                    cropTexture.SetData(data);

                    result[x, y] = cropTexture;
                }
            }

            originalTexture.Dispose();

            return result;
        }

    }
}
